--[[
Rewritten from Incursion Guild Alts, originally by Zorlan.  See there for the
full tortured history.

]]

local addon = select(2,...)

addon.defaults = {
	profile = {
		enable = true,
		alt_pattern = "[Aa][lL][tT][ ]?%((.-)%)",
		do_rank_tags = false,
		do_rank_tags_officer = false,
		hcd_debug = false,
		debugging_frame = 1,
		hcd_debugging_frame = 2,
		max_chatframe_num = 7,
	}
}

--[[
Displaying guild ranks and corresponding names:

  /run for i=1, GuildControlGetNumRanks() do print(i-1, GuildControlGetRankName(i)) end

GetGuildRosterInfo and similar functions return 0-based numbers, thus the -1 above.
XXX This needs to be configurable via the options window.
]]
addon.rank_letter_map = {
	[0] = 'O',  -- GM
	[1] = 'O',  -- leader
	[2] = 'O',  -- officer
	[3] = 'B',  -- raid boss
	[4] = 'R',  -- raider
	[5] = 'I',  -- initiate
	[6] = 'O',  -- alt officer
	[7] = 'M',  -- member
	[8] = 'a',  -- alt
	[9] = 'm',  -- F&F
}

addon.options = {
	name = "Ouro Guild Alts",
	type = 'group',
	childGroups = 'tab',
	handler = addon,    -- functions listed as strings called as addon:func
	get = "GetOption",
	set = "SetOption",
	args = {
		options = {
			name = "Options",
			desc = "General options",
			type = 'group',
			args = {
				header1 = {
					name = "General",
					type = 'header',
					cmdHidden = true,
					order = 1,
				},
				enable = {
					name = "Enable",
					desc = "Use this addon",
					type = 'toggle',
					arg  = "ToggleEnable",
					order = 5,
				},
				rescan = {
					name = "Rescan Guild Roster",
					desc = "Reloads the list of alts from the guild roster",
					type = 'execute',
					func = "LoadGuild",
					order = 10,
				},
				spacer1 = {
					name = (""):rep(50),
					type = 'description',
					cmdHidden = true,
					order = 14,
				},
				do_rank_tags = {
					name = "Use rank tags",
					desc = "Add single-letter rank tags to names (currently in /raid only).  Almost certainly useless for your guild until I make it more configurable.  Sorry.",
					type = 'toggle',
					arg  = "UpdateRankTagging",
					order = 15,
				},
				do_rank_tags_officer = {
					name = "Tags only when officer",
					desc = "Use rank tags only when you are a guild officer ('officer' defined as being able to /ginvite)",
					type = 'toggle',
					arg  = "UpdateRankTagging",
					disabled = function(i) return not i.handler.db.profile.do_rank_tags end,
					order = 16,
				},
				header2 = {
					name = "Advanced",
					type = 'header',
					cmdHidden = true,
					order = 50,
				},
				debug = {
					name = "Toggle debugging",
					desc = "This is a metric assload of text output, especially if HCD is active",
					type = 'toggle',
					arg  = "HookChat",
					order = 55,
				},
				hcd_debug = {
					name = "Toggle HCD",
					desc = "You really want to put this in its own window. |cffff0000REQUIRES /reload AFTER CHANGING.|r",
					type = 'toggle',
					order = 56,
				},
				print_chatframes = {
					name = "Print Chatframe Numbers",
					desc = "Print each chat window number in its frame, for easy reference in the next slider options",
					type = 'execute',
					--func = print_chatframes filled in below
					order = 57,
				},
				debugging_frame = {
					name = "Debugging Output Chatframe",
					desc = "Which chat window to use for normal debug output",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					order = 58,
				},
				hcd_debugging_frame = {
					name = "HCD Debugging Output Chatframe",
					desc = "Which chat window to use for HCD output. |cffff0000REQUIRES /reload AFTER CHANGING.|r",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					order = 59,
				},
				max_chatframe_num = {
					name = "Max Chatframe",
					desc = "Highest chatframe number to hook",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					arg  = "HookChat",
					order = 60,
				},
				alt_pattern = {
					name = "Pattern",
					desc = "Lua pattern to detect and extract names out of the player and officer note fields (see HOWTO.txt for more)",
					type = 'input',
					order = 65,
				},
			},
		},
		--profiles =   filled in OnInit, order 200
	}
}


-----------------------------------------------------------------------------
-- other locals
local guildname, isofficer
local state = setmetatable({
	me = UnitName("player"),
	do_rank_tags = false,
	orig = DEFAULT_CHAT_FRAME.AddMessage, -- on the assumption that all AddMessage's are the same
	channel_match_pattern = false,
	extract_name_and_target = false,
	hcd_debugging_frame = false,
	debugging = function() return addon.db.profile.debug end,
}, {__index = _G})

-- First argument may optionally be a specific chatframe to use, or a boolean
-- overriding the current debug setting.
state.dprint = function(...)
	local t = "<D>"
	local frame
	local do_it = addon.db.profile.debug
	local first = ...
	if type(first) == "table" and first.AddMessage then
		frame, first = first, 2
	elseif type(first) == "boolean" then
		frame, first = addon.db.profile.debugging_frame, 2
		do_it = do_it or first
	else
		frame, first = addon.db.profile.debugging_frame, 1
	end
	if not do_it then return end
	if type(frame) ~= "table" then
		frame = _G["ChatFrame"..frame]
	end
	for i = first, select('#',...) do
		local x = select(i,...)
		t = t .. " " .. tostring(x)   -- take that, garbage collector...
	end
	state.orig(frame,t,1,0,0)  -- specifically not tailcalling
end

local function print_chatframes()
	for i = 1, NUM_CHAT_WINDOWS do
		local cf = _G["ChatFrame"..i]
		if not cf then break end
		state.orig(cf,"This is frame number " .. i .. ".",1,1,1)
	end
end
addon.options.args.options.args.print_chatframes.func = print_chatframes


-- Deal with Prat's various possibilities.  If it's not in use, then do our
-- own chat hooks.  If their alt module is fired up, then add our markup pattern
-- (none of its defaults will match).  Otherise hook into their message format
-- function.
--
-- Returns:
--   "normal" if we should keep doing our thing
--   "bail" if we should leave maps in place but not change chat
--   "cleaup" if we should bail and also clear our own memory
--
--local DO = 0
local function handle_prat()
	if not Prat then return "normal" end
	local mod = Prat.Addon:GetModule("AltNames", --[[silent=]]true)
	if not mod then return "normal" end
	if mod:IsEnabled() then
		local addalt = mod.addAlt
		local pattern = "(%S+)%s+" .. addon.db.profile.alt_pattern
		function mod:addAlt(str)
			local alt,main = str:match(pattern)
			if main then return addalt(self, alt..' '..main)
			else return addalt(self,str)  -- not one of ours
			end
		end
		return "cleanup"
	end
	local function hook(...)
		--if DO == 2 then return end
		local e,m = ...
		local player = m.PLAYERLINK
		--print(e, "player is", player, "escaped is", player:gsub('\124', '\124\124'))
		if player ~= state.me or true then
			--print('replacing', player)
			local mainname = addon.alts_to_main[player]
			--print('main is', mainname)
			if mainname and mainname ~= "" then
				m.PLAYER = ("%s (%s)"):format(player, mainname)
				--print('m.player now', m.PLAYER)
				--DO = DO + 1
			end
		end
	end
	Prat.RegisterChatEvent("OuroGuildAlts", "Prat_FrameMessage", hook)
	--[==[
	Prat.RegisterChatEvent("OuroGuildAlts", "Prat_PreAddMessage", function(...)
		if DO == 2 then return end
		local e,m,_,_,out = ...
		print(e, out:gsub('\124', '\124\124'))
				if DO == 1 then DO = DO + 1 end
	end)
	]==]
	return "bail"
end


-----------------------------------------------------------------------------
addon = LibStub("AceAddon-3.0"):NewAddon(addon, "OuroGuildAlts",
		        "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0")

function addon:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("OuroGuildAltsDB", self.defaults, --[[Default=]]true)

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		self.options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
	end

	LibStub("AceConfig-3.0"):RegisterOptionsTable("OuroGuildAlts", self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("OuroGuildAlts", "Ouro Guild Alts")
	self:RegisterChatCommand("oga", "OnChatCommand")

	self.OnInitialize = nil
end


function addon:FriendsFrame_Pause()
	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE")
end
function addon:FriendsFrame_Resume()
	FriendsFrame:RegisterEvent("WHO_LIST_UPDATE")
end


function addon:OnEnable()
	self:RegisterEvent("PLAYER_LOGOUT")
	if not self.db.profile.enable then
		return self:Disable()
	end

	if self.prat == "cleanup" then
		return self:Print("ERROR: After turning off Prat modules, you will need to relog or /reload to get this addon working.")
	end

	-- Glizzard's code dispatches a GuildRoster() call 15 seconds after load.
	-- These will be waiting, but sometimes still don't catch it.  Don't try
	-- running them until the same 15 seconds have passed.
	AutoCompleteInfoDelayer:HookScript("OnFinished",
		function()
			self:RegisterEvent("GUILD_ROSTER_SHOW", "LoadGuild")
			self:RegisterEvent("GUILD_ROSTER_UPDATE", "LoadGuild")
		end)
end

function addon:OnDisable()
	self:UnhookChat()
end

function addon:ToggleEnable()
	if self.db.profile.enable then
		self:Enable()
		GuildRoster()
	else
		self:Disable()
	end
end

function addon:PLAYER_LOGOUT()
	--self.db.profile.debug = nil
end


do
	local mt = {
		__index =
			function(t,name)
				state.dprint("rank_lookup starting", name)
				local gn,gr,rank = GetGuildInfo(name)
				state.dprint(gn, gr, rank, "and stored guildname is", guildname)
				if gn ~= guildname then return nil end
				rank = addon.rank_letter_map[rank]
				if not rank then return nil end
				t[name] = rank
				state.dprint("rank_lookup returning", name, rank)
				return rank
			end
	}
	function addon:UpdateRankTagging()
		state.do_rank_tags = self.db.profile.do_rank_tags
		if self.db.profile.do_rank_tags_officer then
			state.do_rank_tags = state.do_rank_tags and isofficer
		end
		if state.do_rank_tags and not self.rank_lookup then
			self.rank_lookup = setmetatable({}, mt)
		end
	end
end


function addon:SetOption (info, value)
	local name = info[#info]
	self.db.profile[name] = value
	local arg = info.arg
	if arg then self[arg](self) end
end

function addon:GetOption (info)
	local name = info[#info]
	return self.db.profile[name]
end


function addon:OnChatCommand (input)
	if not input or input:trim() == "" then
		LibStub("AceConfigDialog-3.0"):Open("OuroGuildAlts")
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(self, "oga", "OuroGuildAlts", input)
	end
end


function addon:HookChat ()
	self:UnhookChat()

	self.prat = self.prat or handle_prat()
	handle_prat = nil
	if self.prat == "bail" then
		return
	elseif self.prat == "cleanup" then
		self.rank_letter_map = nil
		self.alts_to_main = nil
		self.rank_lookup = nil
		return
	end

	state.channel_match_pattern = "|Hchannel:(.-)|h"
	if CHAT_TIMESTAMP_FORMAT == nil then
		state.channel_match_pattern = "^" .. state.channel_match_pattern
	end
	state.dprint("channel pattern:", state.channel_match_pattern)

	if not self.AddMessage then
		local func_def
		if self.db.profile.hcd_debug then
			state.hcd_debugging_frame = _G["ChatFrame" .. self.db.profile.hcd_debugging_frame]
			func_def = self.func_def:gsub("%-%-<HCD",""):gsub("%-%->HCD","")
		else
			func_def = self.func_def:gsub("%-%-<HCD.-%-%->HCD","")
		end
		local func,err = loadstring(func_def, "Ouro Guild Alts")
		if func then
			setfenv (func, state)
			self.AddMessage = func()   -- don't pcall this for now
			self.func_def = nil
		else
			self:Print("ERROR: Could not create chat hook functions, giving up.")
			return self:Disable()
		end
	end

	for i = 1, self.db.profile.max_chatframe_num do
		local cf = _G["ChatFrame"..i]
		if not cf then break end
		self:RawHook (cf, "AddMessage")
	end
	self.hooked = true
	if state.debugging() then
		_G.OGA = self
	end
end

function addon:UnhookChat()
	self:UnhookAll()
	self.hooked = nil
end


function addon:LoadGuild ()
	--self:FriendsFrame_Pause()

	local nAlt   = 0
	local nGuild = GetNumGuildMembers(true) or 0

	-- Try and compensate for laggy-ass servers.
	if nGuild == 0 then
		self.lagloop = self.lagloop and (self.lagloop - 1) or 3
		if self.lagloop > 0 then
			return GuildRoster()   -- dispatch a request
		else
			self:Print("ERROR: Repeatedly failed to retrieve guild roster, giving up.")
			return self:Disable()
		end
	end
	self.lagloop = nil

	-- Build the map
	self.alts_to_main = {}
	local pat = self.db.profile.alt_pattern
	for i = 1, nGuild do
		local cleannote
		local name, rank, rankIndex, level, class, zone, note, officernote, online, status = GetGuildRosterInfo(i)

		if officernote and officernote ~= "" then
			cleannote = officernote:match(pat)
		end
		if (not cleannote) and note and note ~= "" then
			cleannote = note:match(pat)
		end
		if cleannote and cleannote ~= "" and name:lower() ~= cleannote:lower() then
			self.alts_to_main[name] = cleannote
			nAlt = nAlt + 1
		end
	end

	self:Printf("Found %d guild members, %d of whom are labeled as alts.", nGuild, nAlt)
	guildname = GetGuildInfo("player")
	isofficer = not not CanGuildInvite()
	self.rank_lookup = nil
	self:UpdateRankTagging()

	--self:FriendsFrame_Resume()
	self:UnregisterEvent("GUILD_ROSTER_SHOW")
	self:UnregisterEvent("GUILD_ROSTER_UPDATE")

	if not self.hooked then
		self:HookChat()
	end
end


--[[
	As of 3.0, raw channel text looks like this:
		|Hchannel:6|h[6. foo]|h |Hplayer:Person:linenumber|h[Person]|h: hello
	and displays like
		[6. foo] [Person]: hello
			The channel 'numbers' of other channels are:  "GUILD" "o" "Party"/"party" "raid"
			Amusingly, "Party" is for normal members and "party" is for the leader.  The raid
			leader chat is already picked up by "raid".
			For system login messages, the ":linenumber" part is not present:
				|Hplayer:Person|h[Person]|h has come online.

	Colorized raid line looks like
	|Hchannel:RAID|h[Raid]|h |Hplayer:Me:linenumber:RAID|h[|cffxxxxxxMe|r]|h: hello
]]


-- Returns (1) the sender's name and (2) the pattern to replace in chat.  If
-- class colors are in use, (2) will be the colorized version of (1), else just
-- a copy of (1).
function state.extract_name_and_target (msg)
	local n, rest = msg:match("|Hplayer:(.-)[:|](.-)|h[: ]")
	if not n then state.dprint(me=='Farmbuyer', "n error on", msg:gsub("\124","\124\124")) return end
	if not rest then state.dprint(me=='Farmbuyer', "rest error on", (n or "<n nil>"):gsub("\124","\124\124"), "from", msg:gsub("\124","\124\124")) return end
	local c = rest:match("|h%[(|cff......)")
	if c then
		c = c .. n .. "|r"
		return n, c
	else return n, n end
end

--[===[
function addon:AddMessage (frame, msg, ...)
	if not self.db.profile.enable then
		-- We shouldn't even be here today!
		self:Disable()
		return self.hooks[frame].AddMessage(frame,msg,...)
	end

	local channel = msg:match(channel_match_pattern)
	channel = channel and channel:lower()
	--XXX HCD
	-- These lines will finish with 4 numbers:  the 3 values fed to the call
	-- to supply a default color, plus the replacement count returned from gsub.
	--dprint(ChatFrame2, self:GetName() .. "<" .. (channel or "") .. ">ORIG:  " .. msg:gsub("\124", "\124\124"), ...)

	if channel == "guild" or channel == "officer" or channel == "party" or channel == "raid"
	   --or (channel == nil and msg:find("|h has come online."))   -- triggers on Read ID logins also
	   or (channel == nil and msg:find("|Hplayer.*|h has come online."))
	then
		local sender, replace_target = extract_name_and_target(msg)

		--XXX HCD
		if sender == me
		   --and not debugging()
		then
			return orig(frame,msg,...)
		end

		-- Try to look it up
		local mainname = self.alts_to_main[sender]
		local replacement

		if mainname then
			replacement = replace_target .. " (" .. mainname .. ")"
		end

		if do_rank_tags and (channel == "raid"
			                 or debugging()
							)
		then
			local r = self.rank_lookup[sender]
			dprint("rank lookup on "..sender.." found "..(r or "nil"))
			if r then
				replacement = (replacement or replace_target) .. ';' .. r
			end
		end

		if replacement then
			--XXX HCD
			--dprint(ChatFrame2, "target: "..(replace_target:gsub("\124","\124\124")))
			--dprint(ChatFrame2, "replacement: "..(replacement:gsub("\124","\124\124")))
			msg = msg:gsub("|h%["..replace_target.."%]|h",
			               "|h%["..replacement.."%]|h")
		end

	end

	return orig(frame,msg,...)
end
]===]


addon.func_def = [=[
return function (self, frame, msg, ...)
	if not self.db.profile.enable then
		-- We shouldn't even be here today!
		self:Disable()
		return self.hooks[frame].AddMessage(frame,msg,...)
	end

	local channel = msg:match(channel_match_pattern)
	channel = channel and channel:lower()
	--<HCD
	dprint(hcd_debugging_frame, frame:GetName() .. "<" .. (channel or "") .. ">:  " .. (msg:gsub("\124", "\124\124")), ...)
	-->HCD

	if channel == "guild" or channel == "officer" or channel == "party" or channel == "raid"
	   --or (channel == nil and msg:find("|h has come online."))   -- triggers on Real ID logins also
	   or (channel == nil and msg:find("|Hplayer.*|h has come online."))
	then
		local sender, replace_target = extract_name_and_target(msg)

		if sender == me
			--<HCD
			and not debugging()
			-->HCD
		then
			return orig(frame,msg,...)
		end

		-- Try to look it up
		local mainname = self.alts_to_main[sender]
		local replacement

		if mainname then
			replacement = replace_target .. " (" .. mainname .. ")"
		end

		if do_rank_tags and (channel == "raid"
			                 or debugging()
							)
		then
			local r = self.rank_lookup[sender]
			dprint("rank lookup on "..sender.." found "..(r or "nil"))
			if r then
				replacement = (replacement or replace_target) .. ';' .. r
			end
		end

		if replacement then
			--<HCD
			dprint(hcd_debugging_frame, "target: "..(replace_target:gsub("\124","\124\124")))
			dprint(hcd_debugging_frame, "replacement: "..(replacement:gsub("\124","\124\124")))
			-->HCD
			msg = msg:gsub("|h%["..replace_target.."%]|h",
			               "|h%["..replacement.."%]|h")
		end

	end

	return orig(frame,msg,...)
end
]=]

-- vim:noet
